% octave 

% $Id: main.m,v 1.21 2013/06/19 09:26:04 mechanoid Exp mechanoid $
%      
%    SVM/SMO
%       
%       OS : GNU/Linux 3.9.3-1-ARCH 
% COMPILER : GNU Octave, version 3.6.4
%
%   AUTHOR : Evgeny S. Borisov
%
%  Glushkov Institute of Cybernetics
%  National Academy of Sciences of Ukraine
% 
%    http://www.mechanoid.kiev.ua
%  e-mail : nn@mechanoid.kiev.ua
% 


% - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - 
% data

% points1 = load('../data/class_0.txt');
% points2 = load('../data/class_1.txt');

% points1 = load('../data/class_0.txt');
% points2 = load('../data/class_2.txt');

% points1 = load('../data/class_0.txt');
% points2 = load('../data/class_3.txt');

% points1 = load('../data/class_1.txt');
% points2 = load('../data/class_2.txt');

% points1 = load('../data/class_1.txt');
% points2 = load('../data/class_3.txt');

points1 = load('../data/class_2.txt');
points2 = load('../data/class_3.txt');

% points1 = load('../data/circ_0.txt');
% points2 = load('../data/circ_1.txt');


scale = 1/100 ;
inputs = [ points1' points2' ] * scale ;

global X = inputs ;
global sp = size(points1,1);
global sn = size(points2,1);
global Y = [ ones(sp,1) ; -1 * ones(sn,1) ] ; % m x 1

% - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - 
%  parameters

global training_set_size = size(inputs,2);

global errorCache = zeros(training_set_size,1) ;
global C = 10.0 ;
global tau = 0.01 ;
global epsilon = 0.01;

global w0 = 0.0 ;
global lambda = zeros(training_set_size,1) ;

global sigma = 0.5 ; % = 1 / ( 2 * sig^2 ) or 1/number of features


% - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - 
% kernel
function retval= kernelGauss(s,x)
   global sigma;
   retval = exp( ( - sum( (s - x).^2 ) ) .* sigma )' ; 
endfunction

function retval= kernelLinear(s,x)
   retval = s' * x ; 
endfunction
 
function retval= kernel(s,x)
   retval = kernelGauss(s,x);
%   retval = kernelLinear(s,x);
endfunction



% - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - 
function retval = svm(a)
   global lambda ;
   global w0 ;
   global X ;
   global Y ;

   retval = sum( kernel(X,a) .* Y .* lambda ) .- w0 ; 
endfunction


 
% = = = = = = = = = = = = = = = = = = = = = = = = = = = = = = = = = = = = = = = = 
% SMO 

% - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - 
function retval = randomSort(arr)
   retval = arr;  

   s = size(retval,2);
   if(s<2), return; end;

   j = 1 ;
   while true,
      % get random indeces
	  i1 = round(rand()*s);
	  i2 = round(rand()*s);

	  if( i1<1 || i2<1 || i1 == i2 ), continue; end;

	  % swap elements
	  b = retval(i1);
	  retval(i1) = retval(i2);
	  retval(i2) = b;

	  j+=1;
	  if(j>s), break; end;
   end;

endfunction



% - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - 
function retval = getError(j)
   global errorCache;
   global lambda;
   global X;
   global Y;
   global C;

   if( 0 < lambda(j) && lambda(j) < C ),
	  retval = errorCache(j);
	  return;
   end;

   retval = svm(X(:,j)) - Y(j) ; 

endfunction


% - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - 
function retval = updateErrorCache(i_2,i_1,t_2,t_1,dw),
   global training_set_size;
   global errorCache;
   global lambda;
   global C;
   global X;
 
   for k = 1:training_set_size,
	  if( 0 < lambda(k) && lambda(k) < C ),
		 K_1k = kernel(X(:,i_1),X(:,k)) ;
		 K_2k = kernel(X(:,i_2),X(:,k)) ;
		 errorCache(k) += t_1*K_1k + t_2*K_2k - dw ;
	  end;
   end ;
   errorCache(i_1) = 0;
   errorCache(i_2) = 0;

endfunction

% - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - 
function retval = takeStep(i_1,i_2),
   global errorCache;
   global lambda;
   global epsilon;
   global w0;
   global X;
   global Y;
   global C;

   retval = 0;

   if( i_1 == i_2 ), return ; end ;

   lambda_1 = lambda(i_1);
   lambda_2 = lambda(i_2);

   y_1 = Y(i_1);
   y_2 = Y(i_2);

   if( y_2 == y_1 ),
	  gama = lambda_1+lambda_2; 
	  if( gama > C ),
			L = gama - C;
			H = C;
		 else
			L = 0;
			H = gama;
	  end;
   else
	  gama = lambda_1-lambda_2; 
	  if( gama > 0 ),
			L = 0;
			H = C - gama;
		 else
			L = -gama;
			H = C;
	  end;
   end; 

   if( L == H ), return; end;

   E_1 = getError(i_1);
   E_2 = getError(i_2);

   K_11 = kernel(X(:,i_1),X(:,i_1)) ;
   K_12 = kernel(X(:,i_1),X(:,i_2)) ; 
   K_22 = kernel(X(:,i_2),X(:,i_2)) ;

   eta = 2*K_12 - K_11 - K_22 ;

   if( eta < 0 ),

	  lambda_2_new = lambda_2 - y_2 * (E_1 - E_2) / eta ;

      if( lambda_2_new < L ), 
		 lambda_2_new = L;
      elseif( lambda_2_new > H ), 
		 lambda_2_new = H;
	  end;
   
   else

	  c1 = eta/2;
	  c2 = y_2*(E_1-E_2) - eta*lambda_2;
	  L_obj = c1 * L * L + c2 * L ;
	  H_obj = c1 * H * H + c2 * H ;

	  if( L_obj > H_obj + epsilon ), 
		 lambda_2_new = L ;
      elseif( L_obj < H_obj - epsilon ), 
		 lambda_2_new = H ;
	  else
		 lambda_2_new = lambda_2;
	  end;

   end;

   ee = 1e-8 ;
   if(lambda_2_new < ee),
	  lambda_2_new = 0 ;
   elseif (lambda_2_new > C - ee),
	  lambda_2_new = C ;
   end;

   if( abs(lambda_2_new - lambda_2) < epsilon * ( lambda_2_new + lambda_2 + epsilon) ),return; end;

   s = y_2*y_1 ;

   lambda_1_new = lambda_1 + s*(lambda_2 - lambda_2_new ) ;
   if (lambda_1_new < 0),
	  lambda_2_new += s*lambda_1_new ;
	  lambda_1_new = 0;
   elseif (lambda_1_new > C ),
	  lambda_2_new += s*(lambda_1_new - C) ;
	  lambda_1_new = C;
   end;
   
   dl_1 = lambda_1_new - lambda_1 ;
   dl_2 = lambda_2_new - lambda_2 ;

   b1 = E_1 + y_1*dl_1*K_11 + y_2*dl_2*K_12 + w0 ;
   b2 = E_2 + y_1*dl_1*K_12 + y_2*dl_2*K_22 + w0;

   if ( 0 < lambda_1_new && lambda_1_new < C ),
	  w0_new = b1 ;
   elseif ( 0 < lambda_2_new && lambda_2_new < C ),
	  w0_new = b2 ;
   else
	  w0_new = (b1+b2)/2 ;
   end;

   t_2 = y_2*dl_2 ;
   t_1 = y_1*dl_1 ;
   dw = w0_new - w0 ;

   updateErrorCache(i_2,i_1,t_2,t_1,dw) ;

   w0 = w0_new;
   lambda(i_2) = lambda_2_new;
   lambda(i_1) = lambda_1_new;

   retval = 1;

endfunction;


% - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - 
function retval = examineExample(i_2)
   global training_set_size;
   global errorCache;
   global lambda;
   global tau;
   global Y;
   global C;

   retval = 0;

   E_2 = getError(i_2) ;
   r = E_2 * Y(i_2) ;

   kkt = ( ( r < -tau ) && (lambda(i_2) < C ) ) || ( ( r >  tau ) && (lambda(i_2) > 0 ) ) ;

   if ( !kkt ), return ; end;

   % search second lambda index
   idx = [] ;
   for i = 1:training_set_size,
	  if( 0 < lambda(i) && lambda(i) < C ),
		 idx = [ idx , i ] ;
	  end;
   end ;

   % search second lambda index
   if( size(idx)>0 ),
	  % try max E
	  i_1 = 1 ;
	  Emax = 0.0 ; 
	  for i = idx,
		 E = abs ( E_2 - errorCache(i) ) ;
		 if( Emax < E ), Emax = E; i_1 = i; end;
	  end ;
	  retval = takeStep(i_1,i_2); 
	  if(retval>0), return; end;


	  % try iterating through non-bound examles
	  idxr = randomSort(idx) ;
	  for i_1 = idxr,
		 retval = takeStep(i_1,i_2);
		 if( retval > 0 ), return; end;
	  end ;

   end;

   % try iterating through the entire traning set
   idx = 1:training_set_size ;
   idxr = randomSort(idx) ;
   for i_1 = idxr,
	  retval = takeStep(i_1,i_2);
      if( retval > 0 ), return; end;
   end ;

endfunction


% - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - 
function  training()
   global training_set_size;
   global lambda;
   global C;
	
   numChanged = 0 ;
   examineAll = 1 ; 

   while ( numChanged > 0 || examineAll > 0 ),

	  numChanged = 0 ;

	  if( examineAll > 0 ), 
		 for i_2 = 1:training_set_size,
			numChanged += examineExample(i_2) ; 
		 end ;
		 examineAll = 0 ;
	  else
		 for i_2 = 1:training_set_size,
			if( 0 < lambda(i_2) && lambda(i_2) < C ),
			  numChanged += examineExample(i_2) ; 
			end;
		 end ;
		 if (numChanged < 1), examineAll = 1 ; end;
	  end;

   end;

endfunction

% = = = = = = = = = = = = = = = = = = = = = = = = = = = = = = = = = = = = = = = = 
% plot data

% - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - 
function plotTrainingSet()
   global training_set_size;
   global X;
   global Y;

   Xdp = [];
   Xdn = [];

   for j = 1:training_set_size,
	  if ( Y(j) > 0 ),
		 Xdp = [ Xdp , X(:,j) ];
	  else
		 Xdn = [ Xdn , X(:,j) ];
	  end;

   end;

   clf ;
   plot( Xdp(1,:) , Xdp(2,:) , '.g' )  ;
   axis([0 1 0 1]) 
   axis("off") 
   print('result/1-xpos.png','-dpng','-S350,350' ) ;

   clf ;
   plot( Xdn(1,:) , Xdn(2,:) , '.g' )  ;
   axis([0 1 0 1]) 
   axis("off") 
   print('result/2-xneg.png','-dpng','-S350,350' ) ;

endfunction

% - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - 
function plotSupport()
   global training_set_size;
   global lambda;
   global X;
   global Y;
   global C;

   % support vectors
   Xs = [];  % 0 < lambda < C
   Xsm = []; % lambda = C

   % training set vectors
   Xdp = []; % positive 
   Xdn = []; % negative 

   s_count = 0;

   % sort data
   for j = 1:training_set_size,

	  if( lambda(j) > 0 ), % support vectors
		 if( lambda(j) < C ), 
			Xs = [ Xs , X(:,j) ];
		 else
			Xsm = [ Xsm , X(:,j) ];
		 end;
		 s_count += 1;
	  end;

	  if ( Y(j)>0 ), % positive 
		 Xdp = [ Xdp , X(:,j) ];
	  else % negative 
		 Xdn = [ Xdn , X(:,j) ];
	  end;

   end ;

   s_count

   if(s_count<1),return ;end;

   clf ;
   hold on ; 
   
   % plot training set vectors
   plot( Xdp(1,:) , Xdp(2,:) , '.g' )  ;
   plot( Xdn(1,:) , Xdn(2,:) , '.b' )  ;
   if(size(Xs)>0),plot( Xs(1,:) , Xs(2,:) , 'o5' ) ; end ;
   if(size(Xsm)>0),plot( Xsm(1,:) , Xsm(2,:) , 'or' ) ; end ;
   

   axis([0 1 0 1]) 
   axis("off") 
   print('result/3-support.png','-dpng','-S450,450' ) ;

endfunction



% - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - 
function makeTest()
   global training_set_size;
   global X;

   clf ;
   hold on ; 
   plot( X(1,:) , X(2,:) , '.b' )  ;
   axis([0 1 0 1]) 
   axis("off") 
   print('result/4-test.png','-dpng','-S350,350' ) ;

   Xdp = [];
   Xdn = [];

   for j = 1:training_set_size,
	  res = svm( X(:,j) );

	  if ( res > 0 ),
		 Xdp = [ Xdp , X(:,j) ];
	  else
		 Xdn = [ Xdn , X(:,j) ];
	  end;

   end;

   clf ;
   hold on ; 
   plot( Xdp(1,:) , Xdp(2,:) , '.r' )  ;
   plot( Xdn(1,:) , Xdn(2,:) , '.g' )  ;
   axis([0 1 0 1]) 
   axis("off") 
   print('result/5-res.png','-dpng','-S350,350' ) ;

endfunction


% - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - 


training();
 
save('-text','result/lambda.txt', 'lambda') ;
save('-text','result/w0.txt', 'w0') ;
 
plotTrainingSet();
plotSupport();

makeTest();

